
/* Chrysalide - Outil d'analyse de fichiers binaires
 * token.c - bribes de recherche textuelle
 *
 * Copyright (C) 2023 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "token.h"


#include <assert.h>
#include <ctype.h>
#include <stdio.h>


#include "token-int.h"
#include "tokens/nodes/plain.h"
#include "../../../common/cpp.h"
#include "../../../core/logs.h"



/* ------------------------- CIBLAGE DES SEQUENCES D'OCTETS ------------------------- */


/* Initialise la classe des bribes de recherche textuelle. */
static void g_bytes_token_class_init(GBytesTokenClass *);

/* Initialise une instance de bribe de recherche textuelle. */
static void g_bytes_token_init(GBytesToken *);

/* Supprime toutes les références externes. */
static void g_bytes_token_dispose(GBytesToken *);

/* Procède à la libération totale de la mémoire. */
static void g_bytes_token_finalize(GBytesToken *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Affiche un motif de recherche au format texte. */
static void g_bytes_token_output_to_text(const GBytesToken *, GScanContext *, int);

/* Affiche un motif de recherche au format JSON. */
static void g_bytes_token_output_to_json(const GBytesToken *, GScanContext *, const sized_string_t *, unsigned int, int);



/* ---------------------------------------------------------------------------------- */
/*                           CIBLAGE DES SEQUENCES D'OCTETS                           */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une bribe de recherche textuelle. */
G_DEFINE_TYPE(GBytesToken, g_bytes_token, G_TYPE_SEARCH_PATTERN);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des bribes de recherche textuelle.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_bytes_token_class_init(GBytesTokenClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GSearchPatternClass *pattern;           /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_bytes_token_dispose;
    object->finalize = (GObjectFinalizeFunc)g_bytes_token_finalize;

    pattern = G_SEARCH_PATTERN_CLASS(klass);

    pattern->to_text = (output_pattern_to_text_fc)g_bytes_token_output_to_text;
    pattern->to_json = (output_pattern_to_json_fc)g_bytes_token_output_to_json;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token = instance à initialiser.                              *
*                                                                             *
*  Description : Initialise une instance de bribe de recherche textuelle.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_bytes_token_init(GBytesToken *token)
{
    token->root = NULL;
    token->slow = 0;
    token->need_backward = false;

    token->fullword = false;
    token->private = false;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_bytes_token_dispose(GBytesToken *token)
{
    g_clear_object(&token->root);

    G_OBJECT_CLASS(g_bytes_token_parent_class)->dispose(G_OBJECT(token));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_bytes_token_finalize(GBytesToken *token)
{
    G_OBJECT_CLASS(g_bytes_token_parent_class)->finalize(G_OBJECT(token));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token    = encadrement de motif à initialiser pleinement.    *
*                root     = représentation du motif à recherche.              *
*                fullword = limite les correspondances à des mots entiers.    *
*                private  = donne une vocation privée au motif de recherche.  *
*                                                                             *
*  Description : Met en place un gestionnaire de recherche de binaire.        *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_bytes_token_create(GBytesToken *token, GScanTokenNode *root, bool fullword, bool private)
{
    bool result;                            /* Bilan à retourner           */

    result = true;

    token->root = root;
    g_object_ref(G_OBJECT(root));

    token->fullword = fullword;
    token->private = private;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token = encadrement de motif à consulter.                    *
*                                                                             *
*  Description : Indique si seuls des mots entiers sont retenus des analyses. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_bytes_token_target_fullword(const GBytesToken *token)
{
    bool result;                            /* Statut à renvoyer           */

    result = token->fullword;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token = encadrement de motif à consulter.                    *
*                                                                             *
*  Description : Détermine si le gestionnaire est à vocation privée.          *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_bytes_token_is_private(const GBytesToken *token)
{
    bool result;                            /* Statut à renvoyer           */

    result = token->private;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token   = définition de la bribe à enregistrer.              *
*                backend = moteur de recherche à préchauffer.                 *
*                maxsize = taille max. des atomes (mise en commun optimisée). *
*                                                                             *
*  Description : Inscrit la définition d'un motif dans un moteur de recherche.*
*                                                                             *
*  Retour      : Bilan de l'opération à renvoyer.                             *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_bytes_token_enroll(GBytesToken *token, GEngineBackend *backend, size_t maxsize)
{
    bool result;                            /* Statut à retourner          */

    token->need_backward = g_scan_token_node_setup_tree(token->root);

    result = g_scan_token_node_enroll(token->root, backend, maxsize, &token->slow);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token   = définition de la bribe à peaufiner.                *
*                backend = moteur de recherche à préchauffer.                 *
*                                                                             *
*  Description : Récupère les identifiants finaux pour un motif recherché.    *
*                                                                             *
*  Retour      : Bilan de l'opération à renvoyer.                             *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_bytes_token_build_id(GBytesToken *token, GEngineBackend *backend)
{
    bool result;                            /* Statut à retourner          */

    result = g_scan_token_node_build_id(token->root, backend);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token   = définition de la bribe à manipuler.                *
*                matches = suivi des correspondances à consolider.            *
*                params  = paramètres des opérations de validation.           *
*                                                                             *
*  Description : Transforme les correspondances locales en trouvailles.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_bytes_token_check(const GBytesToken *token, GScanBytesMatches *matches, scan_node_check_params_t *params)
{
    match_area_t *area;                     /* Correspondance à valider    */
    match_area_t *next;                     /* Correspondance suivante     */
    vmpa2t pos;                             /* Tête de lecture             */
    const bin_t *byte;                      /* Octet à valider             */

    /* Réinitialisation */

    // TODO: offset

    params->initialized = false;

    params->main_areas = NULL;
    params->main_count = 0;

    params->created_areas = NULL;
    params->created_count = 0;

    params->kept_areas = NULL;
    params->kept_count = 0;

    /* Lancement des analyses */

    g_scan_token_node_check_forward(token->root, params);

    if (token->need_backward)
        g_scan_token_node_check_backward(token->root, params);

    // REMME ? sort_and_filter_pending_matches(matches);

    if (token->fullword)
    {
        for_each_match_area_safe(area, &params->main_areas, next)
        {
            /* Validation de l'octet précédent, s'il existe */
            if (area->start > params->content_start)
            {
                init_vmpa(&pos, area->start - 1, VMPA_NO_VIRTUAL);

                byte = g_binary_content_get_raw_access(params->content, &pos, 1);

                if (isalnum(*byte))
                {
                    del_match_area(area, &params->main_areas);
                    assert(&params->main_count > 0);
                    params->main_count--;
                    continue;
                }

            }

            /* Validation de l'octet suivant, s'il existe */
            if (area->end < params->content_end)
            {
                init_vmpa(&pos, area->end, VMPA_NO_VIRTUAL);

                byte = g_binary_content_get_raw_access(params->content, &pos, 1);

                if (isalnum(*byte))
                {
                    del_match_area(area, &params->main_areas);
                    assert(&params->main_count > 0);
                    params->main_count--;
                    continue;
                }

            }

        }

    }

    g_scan_bytes_matches_set_list(matches, params->main_areas, params->main_count);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : token   = définition de la bribe à consulter.                *
*                index = indice de la combinaison de modificateurs ciblée.    *
*                                                                             *
*  Description : Retrouve l'origine d'une correspondance à partir d'un indice.*
*                                                                             *
*  Retour      : Version humainement lisible de la combinaison gagnante.      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

char *g_bytes_token_get_modifier_path(const GBytesToken *token, size_t index)
{
    char *result;                           /* Combinaison à retourner     */

    if (G_IS_SCAN_TOKEN_NODE_PLAIN(token->root))
        result = g_scan_token_node_plain_get_modifier_path(G_SCAN_TOKEN_NODE_PLAIN(token->root), index);
    else
        result = NULL;

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : pattern = définition de motif à considérer.                  *
*                context = contexte de l'analyse à mener.                     *
*                fd      = canal d'écriture.                                  *
*                                                                             *
*  Description : Affiche un motif de recherche au format texte.               *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_bytes_token_output_to_text(const GBytesToken *pattern, GScanContext *context, int fd)
{
    GScanMatches *matches;                  /* Correspondances établies    */

    if (g_bytes_token_is_private(pattern))
        return;

    matches = g_scan_context_get_full_matches(context, G_SEARCH_PATTERN(pattern));

    if (matches != NULL)
    {
        g_scan_matches_output_to_text(matches, fd);
    
        g_object_unref(G_OBJECT(matches));

    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : pattern = définition de motif à considérer.                  *
*                context  = contexte de l'analyse à mener.                    *
*                padding  = éventuel bourrage initial à placer ou NULL.       *
*                level    = profondeur actuelle.                              *
*                fd       = canal d'écriture.                                 *
*                                                                             *
*  Description : Affiche un motif de recherche au format JSON.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_bytes_token_output_to_json(const GBytesToken *pattern, GScanContext *context, const sized_string_t *padding, unsigned int level, int fd)
{
    GScanMatches *matches;                  /* Correspondances établies    */

    if (g_bytes_token_is_private(pattern))
        return;

    matches = g_scan_context_get_full_matches(context, G_SEARCH_PATTERN(pattern));

    if (matches != NULL)
    {
        g_scan_matches_output_to_json(matches, padding, level, fd);
    
        g_object_unref(G_OBJECT(matches));

    }

}
